@@ -20,7 +20,7 @@
 #include <linux/serial.h>
 #include <linux/tty.h>
 
-#define VERSION_STR "2.10"
+#define VERSION_STR "2.12"
 
 char *progname;
 
@@ -37,6 +37,12 @@ struct serial_type_struct {
 	PORT_16450,	"16450",
 	PORT_16550,	"16550",
 	PORT_16550A,	"16550A",
+#ifdef PORT_CIRRUS
+	PORT_CIRRUS,	"Cirrus",
+#endif
+#ifdef PORT_16650
+	PORT_16650,	"16650",
+#endif
 	PORT_UNKNOWN,	"none",
 	-1,		NULL
 };
@@ -48,7 +54,11 @@ struct serial_type_struct {
 #define CMD_TYPE	5
 #define CMD_BASE	6
 #define CMD_DELAY	7
-#define CMD_CONFIG	8
+#define CMD_WAIT	8
+#define CMD_WAIT2	9
+#define CMD_CONFIG	10
+#define CMD_GETMULTI	11
+#define CMD_SETMULTI	12
 
 #define FLAG_CAN_INVERT	0x0001
 #define FLAG_NEED_ARG	0x0002
@@ -87,7 +97,13 @@ struct flag_type_table {
 	CMD_BASE,	"base",		0,		0,		0, FLAG_NEED_ARG,
 	CMD_BASE,	"baud_base",	0,		0,		0, FLAG_NEED_ARG,
 	CMD_DELAY,	"close_delay",	0,		0,		0, FLAG_NEED_ARG,
+#ifdef ASYNC_CLOSING_WAIT_NONE
+	CMD_WAIT,	"closing_wait",	0,		0,		0, FLAG_NEED_ARG,
+	CMD_WAIT2,	"closing_wait2",0,		0,		0, FLAG_NEED_ARG,
+#endif
 	CMD_CONFIG,	"autoconfig",	0,		0,		0, 0,
+	CMD_GETMULTI,	"get_multiport",0,		0,		0, 0,
+	CMD_SETMULTI,	"set_multiport",0,		0,		0, 0,
 	0,		0,		0,		0,		0, 0,
 };
 	
@@ -155,10 +171,139 @@ void print_flags(struct serial_struct *serinfo,
 		printf("%s", postfix);
 }
 
+#ifdef TIOCSERGETMULTI
+void multiport_usage()
+{
+
+	fprintf(stderr, "\nValid keywords after set_multiport are:\n");
+	fprintf(stderr, "\tport_monitor, port[1-4], mask[1-4], "
+		"match[1-4]\n\n");
+	fprintf(stderr, "All arguments take an numeric argument following "
+		"the keyword.\n");
+	fprintf(stderr, "Use a leading '0x' for hex numbers.\n\n");
+}
+
+void get_multiport(char *device, int fd)
+{
+	struct serial_multiport_struct multi;
+
+	if (ioctl(fd, TIOCSERGETMULTI, &multi) < 0) {
+		perror("Cannot get multiport config");
+		exit(1);
+	}
+	printf("Multiport config for irq %d:\n", multi.irq);
+	printf("\tPort monitor = 0x%x\n", multi.port_monitor);
+	printf("\tPort1 = 0x%x, mask=0x%x, match=0x%x\n", multi.port1,
+	       multi.mask1, multi.match1);
+	printf("\tPort2 = 0x%x, mask=0x%x, match=0x%x\n", multi.port2,
+	       multi.mask2, multi.match2);
+	printf("\tPort3 = 0x%x, mask=0x%x, match=0x%x\n", multi.port3,
+	       multi.mask3, multi.match3);
+	printf("\tPort4 = 0x%x, mask=0x%x, match=0x%x\n", multi.port4,
+	       multi.mask4, multi.match4);
+}
+
+void set_multiport(char *device, int fd, char ***in_arg)
+{
+	char **arg = *in_arg;
+	char *word, *argument;
+	struct serial_multiport_struct multi;
+
+	if (ioctl(fd, TIOCSERGETMULTI, &multi) < 0) {
+		perror("Cannot get multiport config");
+		exit(1);
+	}
+	if (*arg == 0) {
+		multiport_usage();
+		return;
+	}
+	while (*arg) {
+		word = *arg++;
+		if (*arg == 0) {
+			multiport_usage();
+			exit(1);
+		}
+		argument = *arg++;
+		if (strcasecmp(word, "port_monitor") == 0) {
+			multi.port_monitor = atonum(argument);
+			continue;
+		}
+		if (strcasecmp(word, "port1") == 0) {
+			multi.port1 = atonum(argument);
+			continue;
+		}
+		if (strcasecmp(word, "mask1") == 0) {
+			multi.mask1 = atonum(argument);
+			continue;
+		}
+		if (strcasecmp(word, "match1") == 0) {
+			multi.match1 = atonum(argument);
+			continue;
+		}
+		if (strcasecmp(word, "port2") == 0) {
+			multi.port2 = atonum(argument);
+			continue;
+		}
+		if (strcasecmp(word, "mask2") == 0) {
+			multi.mask2 = atonum(argument);
+			continue;
+		}
+		if (strcasecmp(word, "match2") == 0) {
+			multi.match2 = atonum(argument);
+			continue;
+		}
+		if (strcasecmp(word, "port3") == 0) {
+			multi.port2 = atonum(argument);
+			continue;
+		}
+		if (strcasecmp(word, "mask3") == 0) {
+			multi.mask2 = atonum(argument);
+			continue;
+		}
+		if (strcasecmp(word, "match3") == 0) {
+			multi.match2 = atonum(argument);
+			continue;
+		}
+		if (strcasecmp(word, "port4") == 0) {
+			multi.port2 = atonum(argument);
+			continue;
+		}
+		if (strcasecmp(word, "mask4") == 0) {
+			multi.mask2 = atonum(argument);
+			continue;
+		}
+		if (strcasecmp(word, "match4") == 0) {
+			multi.match2 = atonum(argument);
+			continue;
+		}
+		fprintf(stderr, "Unknown keyword %s.\n", word);
+		multiport_usage();
+		exit(1);
+	}
+	if (ioctl(fd, TIOCSERSETMULTI, &multi) < 0) {
+		perror("Cannot set multiport config");
+		exit(1);
+	}
+	*in_arg = arg;
+}
+#else
+void get_multiport(char *device, int fd)
+{
+	printf("Setserial was compiled under a kernel which did not\n");
+	printf("support the special serial multiport configs.\n");
+}
+void set_multiport(char *device, int fd, char ***in_arg)
+{
+	printf("Setserial was compiled under a kernel which did not\n");
+	printf("support the special serial multiport configs.\n");
+}
+#endif
+
 void get_serial(char *device)
 {
 	struct serial_struct serinfo;
 	int	fd;
+	char	buf1[40], buf2[40];
 
 	if ((fd = open(device, O_RDWR|O_NONBLOCK)) < 0) {
 		perror(device);
@@ -178,6 +323,21 @@ void get_serial(char *device)
 		printf("\tBaud_base: %d, close_delay: %d, divisor: %d\n",
 		       serinfo.baud_base, serinfo.close_delay,
 		       serinfo.custom_divisor);
+#ifdef ASYNC_CLOSING_WAIT_NONE
+		if (serinfo.closing_wait == ASYNC_CLOSING_WAIT_INF)
+			strcpy(buf1, "infinte");
+		else if (serinfo.closing_wait == ASYNC_CLOSING_WAIT_NONE)
+			strcpy(buf1, "none");
+		else
+			sprintf(buf1, "%d", serinfo.closing_wait);
+		if (serinfo.closing_wait2 == ASYNC_CLOSING_WAIT_INF)
+			strcpy(buf2, "infinte");
+		else if (serinfo.closing_wait2 == ASYNC_CLOSING_WAIT_NONE)
+			strcpy(buf2, "none");
+		else
+			sprintf(buf2, "%d", serinfo.closing_wait2);
+		printf("\tclosing_wait: %s, closing_wait2: %s\n", buf1, buf2);
+#endif
 		print_flags(&serinfo, "\tFlags: ", "");
 		printf("\n\n");
 	} else if (verbosity==0) {
@@ -208,8 +368,8 @@ void set_serial(char *device, char ** arg)
 	
 
 	if ((fd = open(device, O_RDWR|O_NONBLOCK)) < 0) {
-		if (verbosity==0 && (errno==ENOENT||errno==ENODEV))
-			return;
+		if (verbosity==0 && errno==ENOENT)
+			exit(201);
 		perror(device);
 		exit(201);
 	}
@@ -269,6 +429,26 @@ void set_serial(char *device, char ** arg)
 		case CMD_DELAY:
 			new_serinfo.close_delay = atonum(*arg++);
 			break;
+#ifdef ASYNC_CLOSING_WAIT_NONE
+		case CMD_WAIT:
+			if (!strcasecmp(*arg, "infinite"))
+				new_serinfo.closing_wait = ASYNC_CLOSING_WAIT_INF;
+			else if (!strcasecmp(*arg, "none"))
+				new_serinfo.closing_wait = ASYNC_CLOSING_WAIT_NONE;
+			else
+				new_serinfo.closing_wait = atonum(*arg);
+			arg++;
+			break;
+		case CMD_WAIT2:
+			if (!strcasecmp(*arg, "infinite"))
+				new_serinfo.closing_wait2 = ASYNC_CLOSING_WAIT_INF;
+			else if (!strcasecmp(*arg, "none"))
+				new_serinfo.closing_wait2 = ASYNC_CLOSING_WAIT_NONE;
+			else
+				new_serinfo.closing_wait2 = atonum(*arg);
+			arg++;
+			break;
+#endif
 		case CMD_CONFIG:
 			if (ioctl(fd, TIOCSSERIAL, &new_serinfo) < 0) {
 				perror("Cannot set serial info");
@@ -283,6 +463,21 @@ void set_serial(char *device, char ** arg)
 				exit(1);
 			}
 			break;
+		case CMD_GETMULTI:
+			if (ioctl(fd, TIOCSSERIAL, &new_serinfo) < 0) {
+				perror("Cannot set serial info");
+				exit(1);
+			}
+			get_multiport(device, fd);
+			break;
+		case CMD_SETMULTI:
+			if (ioctl(fd, TIOCSSERIAL, &new_serinfo) < 0) {
+				perror("Cannot set serial info");
+				exit(1);
+			}
+			set_multiport(device, fd, &arg);
+			break;
+			
 		default:
 			fprintf(stderr, "Internal error: unhandled cmd #%d\n", p->cmd);
 			exit(1);
@@ -347,7 +542,16 @@ fprintf(stderr, "\t* port\t\tset the I/O port\n");
 	fprintf(stderr, "\t* close_delay\tset the amount of time (in 1/100 of a\n");
 	fprintf(stderr, "\t\t\t\tsecond) that DTR should be kept low\n");
 	fprintf(stderr, "\t\t\t\twhile being closed\n");
-	
+#ifdef ASYNC_CLOSING_WAIT_NONE
+	fprintf(stderr, "\t* closing_wait\tset the amount of time (in 1/100 of a\n");
+	fprintf(stderr, "\t\t\t\tsecond) that the serial port should wait for\n");
+	fprintf(stderr, "\t\t\t\tdata to be drained while being closed, \n");
+	fprintf(stderr, "\t\t\t\tbefore the reciver is disabled\n");
+	fprintf(stderr, "\t* closing_wait2\tset the amount of time (in 1/100 of a\n");
+	fprintf(stderr, "\t\t\t\tsecond) that the serial port should wait for\n");
+	fprintf(stderr, "\t\t\t\tdata to be drained while being closed, \n");
+	fprintf(stderr, "\t\t\t\tafter the reciver is disabled\n");
+#endif
 	fprintf(stderr, "\t^ fourport\tconfigure the port as an AST Fourport\n");
 	fprintf(stderr, "\t  autoconfigure\tautomatically configure the serial port\n");
 	fprintf(stderr, "\t^ auto_irq\ttry to determine irq during autoconfiguration\n");
@@ -364,6 +568,8 @@ fprintf(stderr, "\t* port\t\tset the I/O port\n");
 	fprintf(stderr, "\t^ split_termios Use separate termios for callout and dailin lines\n");
 #endif	
 	fprintf(stderr, "\t^ hup_notify\tNotify a process blocked on opening a dial in line\n");
+	fprintf(stderr, "\t  get_multiport\tDisplay the multiport configuration\n");
+	fprintf(stderr, "\t  set_multiport\tSet the multiport configuration\n");
 	fprintf(stderr, "\t\t\t\twhen a process has finished using a callout\n");
 	fprintf(stderr, "\t\t\t\tline by returning EAGAIN to the open.\n");
 	fprintf(stderr, "\n");
